package Go.strategy;

import java.util.Enumeration;
import java.util.Vector;

import Go.Move;
import Go.GoPosition;
import Go.GoConstants;
import Go.GobanLocation;

// This class models the Alpha-Beta Pruning algorithm for game trees: it's
// an improvement over the classical Min-Max algorithm
// The idea is to avoid exploration of clearly hopeless branches
// The result is the same as with Min-Max, but much faster (especially
// when we have a good ordering of the to-be-explored nodes)

public class AlphaBetaPruningGoStrategy extends MinMaxGoStrategy /*implements GoStrategy*/ {

  // constructor
  public AlphaBetaPruningGoStrategy(int color,
                                    GoEstimatorInterface estimator) {
    this.color = color;
    this.estimator = estimator;
  }

  // method from GoStrategy interface
  public Move chooseNextMove(GoPosition position) {
    nrVariations = 0;

    // start building the tree, with the current position as root
    GameTreeInterface tree = buildTree(position);

    System.out.println("Computed variations = " + nrVariations);
    return ((GoNodeInfo)tree.getRoot().getContent()).nextMove;
  }

  protected GameTreeInterface buildTree(GoPosition position) {
    int rootLastMoveColor = ((position.turn() == GoConstants.BLACK) ? GoConstants.WHITE : GoConstants.BLACK);
    float rootInitialValue = ((position.turn() == this.color) ? Float.NEGATIVE_INFINITY : Float.POSITIVE_INFINITY);

    GameTreeInterface tree = new GameTree(new GoNodeInfo(rootInitialValue,
                                                         new Move(rootLastMoveColor),
                                                         position));

    alpha = Float.NEGATIVE_INFINITY;
    beta = Float.POSITIVE_INFINITY;

    expandAlphaBeta(tree.getRoot(), position, maxDepth, alpha, beta);
	  return tree;
  }

  protected float expandAlphaBeta(TreeNodeInterface node, GoPosition position, int maxDepth,
                                 float alpha, float beta) {

    GoNodeInfo currentNodeInfo = (GoNodeInfo)(node.getContent());
    GoNodeInfo currentChildNodeInfo = null;
    float value, best = Float.NEGATIVE_INFINITY;

    nrVariations++;

    if ((maxDepth == 0)
         || (estimator.checkPositionState(position) != GoConstants.InProgress)) {
      // the current node is at the maximal depth, so we should estimate it
      // and not expand it any further
      best = estimator.estimatePosition(position,
                                        position.turn());
                                        //position.getLastMove().color);
      return best;
    }


    Enumeration nextMoveCandidateEnumeration = null;

    // If a do-or-die move is necessary, then forget about the rest!
    Move doOrDieMove = BasicAtariGoStrategy.tryBasicMove(position);

    if (doOrDieMove != null) {
      Vector oneElementVector = new Vector();
      oneElementVector.addElement(new GobanLocation(doOrDieMove, position));
      nextMoveCandidateEnumeration = oneElementVector.elements();
    }
    else {
      nextMoveCandidateEnumeration = BasicAtariGoStrategy.contactMoves(position);

      // Select as possible next move just the contact moves
      // Not as a best strategy :-) but as a reasonable filtering method
      /*nextMoveCandidateEnumeration = BasicAtariGoStrategy.extremeAttackDefenseMoves(
                                                                        position);*/
    }

    if (!nextMoveCandidateEnumeration.hasMoreElements())
      nextMoveCandidateEnumeration = position.legalMoves_GL();

    Move currentNextMove = null;
    GoPosition nextPosition = null;

    if (nextMoveCandidateEnumeration.hasMoreElements()) {
      while (nextMoveCandidateEnumeration.hasMoreElements()
             && (best < beta)) {
        currentNextMove = new Move((GobanLocation)nextMoveCandidateEnumeration.nextElement());
        nextPosition = position.positionAfterMove(currentNextMove);

        float initialValue = Float.NEGATIVE_INFINITY;
        TreeNode childNode = new TreeNode(new GoNodeInfo(initialValue,
                                                         currentNextMove,
                                                         nextPosition),
                                          node);

        if (best > alpha) {
          alpha = best;
        }

        value = - expandAlphaBeta(childNode, nextPosition, maxDepth - 1, -beta, -alpha);
        currentChildNodeInfo = ((GoNodeInfo)childNode.getContent());

        if (value > best) {
          best = value;
          currentNodeInfo.nextMove = currentNextMove;
        }
      }
    }
    else {
      // the current node is a leaf, so we should estimate it
      best =  estimator.estimatePosition(position,
                                         position.turn());
                                        //position.getLastMove().color);
    }

    return best;
  }

  // The players's color
  // (we need this in order to know when to minimize  and when to
  // maximize positions values)
  // By default, we consider that "color" player tries to maximize, and
  // his oponent to minimize values
  protected int color;

  // a GoEstimatorInterface used to estimate the value of a given position
  protected GoEstimatorInterface estimator;

  //protected int maxDepth = 1;

  private int nrVariations = 0;

  protected float alpha, beta;
}
