package Go.strategy;

import java.util.Enumeration;

import Go.GoPosition;
import Go.Move;
import Go.GoConstants;

// This class models the classical Min-Max algorithm for making decisions in game trees:
// For each node (corresponding to a position in the game) we choose the maximum value descendant
// ( = next move ) when it's our turn to move and the minimum value when it's opponent's turn  

public class MinMaxGoStrategy implements GoStrategy{

  // constructor
  public MinMaxGoStrategy() {
  }

  public MinMaxGoStrategy(int color, GoEstimatorInterface estimator) {
    this.color = color;
    this.estimator = estimator;
  }

  // method from GoStrategy interface
  public Move chooseNextMove(GoPosition position) {
    return getMoveMinMax(position);
  }

  public GoEstimatorInterface getEstimator() {
    return estimator;
  }

  public void setEstimator(GoEstimatorInterface newEstimator) {
    this.estimator = newEstimator;
  }

  public void setMaxDepth(int newMaxDepth) {
    maxDepth = newMaxDepth;
  }

  public int getMaxDepth() {
    return maxDepth;
  }

  // Create a MinMax tree and choose the best Max move
  protected Move getMoveMinMax(GoPosition position) {
    nrVariations = 0;

    // start building the tree, with the current position as root
    GameTreeInterface tree = buildTree(position);

    System.out.println("Computed variations = " + nrVariations);

    return ((GoNodeInfo)tree.getRoot().getContent()).nextMove;
  }

  // Builds the game tree starting from the current position, then labels the
  // nodes with estimated values
  protected GameTreeInterface buildTree(GoPosition position) {
    int rootLastMoveColor = ((position.turn() == GoConstants.BLACK) ? GoConstants.WHITE : GoConstants.BLACK);
    float rootInitialValue = ((position.turn() == this.color) ? Float.NEGATIVE_INFINITY : Float.POSITIVE_INFINITY);

    GameTreeInterface tree = new GameTree(new GoNodeInfo(rootInitialValue,
                                                         new Move(rootLastMoveColor),
                                                         position));
    if (maxDepth!= GoStrategyConstants.NoDepthLimits)
      expandNode(tree.getRoot(), position, maxDepth);
    else
      expandNodeNoLimits(tree.getRoot(), position);
	 return tree;
  }

  // Choose the maximum value move according to all possible moves (estimated) value
  protected Move chooseMaxMove(GameTreeInterface tree) {
    Move bestMove = null;
    float bestMaxValue = Float.NEGATIVE_INFINITY;
    GoNodeInfo nodeInfo;

    Enumeration e = tree.getRoot().getChildren().elements();
    for (;e.hasMoreElements();) {
      nodeInfo = (GoNodeInfo)((TreeNodeInterface)(e.nextElement())).getContent();

      // this test should be eliminated when we are sure everything is OK
      // with tree labeling
      if (nodeInfo.estimatedValue != GoStrategyConstants.notEvaluated)
        if (nodeInfo.estimatedValue > bestMaxValue) {
          bestMove = nodeInfo.lastMove;
          bestMaxValue = nodeInfo.estimatedValue;
        }
    }
    return bestMove;
  }

  // Choose the minimum value move according to all possible moves (estimated) value
  protected Move chooseMinMove(GameTreeInterface tree) {
    Move bestMove = null;
    float bestMinValue = Float.POSITIVE_INFINITY;
    GoNodeInfo nodeInfo;

    Enumeration e = tree.getRoot().getChildren().elements();
    for (;e.hasMoreElements();) {
      nodeInfo = ((GoNodeInfo)((TreeNodeInterface)(e.nextElement())).getContent());

      // this test should be eliminated when we are sure everything is OK
      // with tree labeling
      if (nodeInfo.estimatedValue != GoStrategyConstants.notEvaluated)
        if (nodeInfo.estimatedValue < bestMinValue) {
          bestMove = nodeInfo.lastMove;
          bestMinValue = nodeInfo.estimatedValue;
        }
    }
    return bestMove;
  }

  // Choose the maximum value move according to all possible moves (estimated) value
  protected float chooseMaxValue(TreeNodeInterface node) {
    float bestMaxValue = Float.NEGATIVE_INFINITY;
    GoNodeInfo nodeInfo;

    Enumeration e = node.getChildren().elements();
    for (;e.hasMoreElements();) {
      nodeInfo = ((GoNodeInfo)((TreeNodeInterface)(e.nextElement())).getContent());

      // this test should be eliminated when we are sure everything is OK
      // with tree labeling
      if (nodeInfo.estimatedValue != GoStrategyConstants.notEvaluated)
        if (nodeInfo.estimatedValue > bestMaxValue) {
          bestMaxValue = nodeInfo.estimatedValue;
        }
    }
    return bestMaxValue;
  }

  // Choose the minimum value from all direct descendants
  protected float chooseMinValue(TreeNodeInterface node) {
    float bestMinValue = Float.POSITIVE_INFINITY;
    GoNodeInfo nodeInfo;

    Enumeration e = node.getChildren().elements();
    for (;e.hasMoreElements();) {
      nodeInfo = ((GoNodeInfo)((TreeNodeInterface)(e.nextElement())).getContent());

      // this test should be eliminated when we are sure everything is OK
      // with tree labeling
      if (nodeInfo.estimatedValue != GoStrategyConstants.notEvaluated)
        if (nodeInfo.estimatedValue < bestMinValue) {
          bestMinValue = nodeInfo.estimatedValue;
        }
    }
    return bestMinValue;
  }

  // Calculates every node's value, according to min-max strategy
  protected void computeTreeNodeValue(TreeNodeInterface node) {
    GoNodeInfo currentNodeInfo = (GoNodeInfo)(node.getContent());
    int turn = currentNodeInfo.turn();

    // if the node is not a leaf, then we'll evaluate it based on it's descendants
    // values, recursively
    if (!(node.isLeaf())) {
      Enumeration e = node.getChildren().elements();
      for (;e.hasMoreElements();) {
        // recursive call of this method for all node's children
        computeTreeNodeValue((TreeNodeInterface)(e.nextElement()));
      }

      // now that we have the estimated values for all direct descendants,
      // we can evaluate the node
      if (this.color == turn) currentNodeInfo.estimatedValue = chooseMaxValue(node);
      else currentNodeInfo.estimatedValue = chooseMinValue(node);
    }

    // if node is a leaf, than we have to estimate it's value
    else {
      currentNodeInfo.estimatedValue = estimator.estimatePosition(currentNodeInfo.position,
                                                                  this.color);
    }
  }

  // expands the argument node adding as it's children all the positions
  // that can result by one legal move and evaluate it in the game context
  private void expandNodeNoLimits(TreeNodeInterface node, GoPosition position) {
    GoPosition currentPosition, nextPosition;
    Move currentNextMove;
    TreeNodeInterface parentNode;
    TreeNodeInterface childNode;

    GoNodeInfo currentNodeInfo = (GoNodeInfo)(node.getContent());
    GoNodeInfo currentChildNodeInfo;

    int turn = currentNodeInfo.turn();
    float initialValue;

    System.out.println(++nrVariations);

    // expand the current node
    currentPosition = position;

    // check if the currentPosition is final or not
    Enumeration legalMovesEnumeration = currentPosition.legalMoves();

    if (legalMovesEnumeration.hasMoreElements()) {
      for (;legalMovesEnumeration.hasMoreElements();) {
        currentNextMove = (Move)legalMovesEnumeration.nextElement();
        nextPosition = currentPosition.positionAfterMove(currentNextMove);

        // check to avoid illegal ko recapture
        parentNode = node.getParent();
        if ((parentNode == null) ||
            ((parentNode != null) &&
             (!(nextPosition.sameAs(((GoNodeInfo)parentNode.getContent()).position))))) {
          initialValue = ((this.color == turn) ? Float.POSITIVE_INFINITY : Float.NEGATIVE_INFINITY);
          childNode = new TreeNode(new GoNodeInfo(initialValue,
                                                  currentNextMove,
                                                  nextPosition),
                                   node);

          expandNodeNoLimits(childNode, nextPosition);

          currentChildNodeInfo = ((GoNodeInfo)childNode.getContent());

          if (this.color == turn) {
            // try to maximize values
            if (currentNodeInfo.estimatedValue != Float.POSITIVE_INFINITY) {
              if (currentChildNodeInfo.estimatedValue > currentNodeInfo.estimatedValue){
                currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
                currentNodeInfo.nextMove = currentNextMove;
              }
            }
            else {
              currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
              currentNodeInfo.nextMove = currentNextMove;
            }
          }
          else {
            // try to minimize values
            if (currentNodeInfo.estimatedValue != Float.NEGATIVE_INFINITY) {
              if (currentChildNodeInfo.estimatedValue < currentNodeInfo.estimatedValue) {
                currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
                currentNodeInfo.nextMove = currentNextMove;
              }
            }
            else {
              currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
              currentNodeInfo.nextMove = currentNextMove;
            }
          }
        }
      }
    }
    else {
      // the current node is a leaf, so we should estimate it
      currentNodeInfo.estimatedValue = estimator.estimatePosition(position,
                                                                  this.color);
    }
  }

  // expands the argument node adding as it's children all the positions
  // that can result by one legal move, but going no deeper than maxDepth
  private void expandNode(TreeNodeInterface node, GoPosition position, int maxDepth) {
    GoPosition currentPosition, nextPosition;
    Move currentNextMove;
    TreeNodeInterface parentNode;
    TreeNodeInterface childNode;

    GoNodeInfo currentNodeInfo = (GoNodeInfo)(node.getContent());
    GoNodeInfo currentChildNodeInfo;

    int turn = currentNodeInfo.turn();
    float initialValue;

    nrVariations++;

    // expand the current node
    currentPosition = position;

    if ((maxDepth == 0) ||
         (estimator.checkPositionState(position) != GoConstants.InProgress)) {
      // the current node is at the maximal depth, so we should estimate it
      // and not expand it any further
      currentNodeInfo.estimatedValue = estimator.estimatePosition(position,
                                                                  this.color);
      return;
    }

    // get the legal next moves set
    Enumeration legalMovesEnumeration = currentPosition.legalMoves();

    if (legalMovesEnumeration.hasMoreElements()) {
      for (;legalMovesEnumeration.hasMoreElements();) {
        currentNextMove = (Move)legalMovesEnumeration.nextElement();
        nextPosition = currentPosition.positionAfterMove(currentNextMove);

          initialValue = ((this.color == turn) ? Float.POSITIVE_INFINITY : Float.NEGATIVE_INFINITY);
          childNode = new TreeNode(new GoNodeInfo(initialValue,
                                                  currentNextMove,
                                                  nextPosition),
                                   node);

          expandNode(childNode, nextPosition, maxDepth - 1);

          currentChildNodeInfo = ((GoNodeInfo)childNode.getContent());

          if (this.color == turn) {
            // try to maximize values
            if (currentNodeInfo.estimatedValue != Float.NEGATIVE_INFINITY) {
              if (currentChildNodeInfo.estimatedValue > currentNodeInfo.estimatedValue){
                currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
                currentNodeInfo.nextMove = currentNextMove;
              }
            }
            else {
              currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
              currentNodeInfo.nextMove = currentNextMove;
            }
          }
          else {
            // try to minimize values
            if (currentNodeInfo.estimatedValue != Float.POSITIVE_INFINITY) {
              if (currentChildNodeInfo.estimatedValue < currentNodeInfo.estimatedValue) {
                currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
                currentNodeInfo.nextMove = currentNextMove;
              }
            }
            else {
              currentNodeInfo.estimatedValue = currentChildNodeInfo.estimatedValue;
              currentNodeInfo.nextMove = currentNextMove;
            }
          }
        }
      }
    else {
      // the current node is a leaf, so we should estimate it
      currentNodeInfo.estimatedValue = estimator.estimatePosition(position,
                                                                  this.color);
    }
  }

  // The players's color
  // (we need this in order to know when to minimize  and when to
  // maximize positions values)
  // By default, we consider that "color" player tries to maximize, and
  // his oponent to minimize values
  protected int color;

  // a GoEstimatorInterface used to estimate the value of a given position
  protected GoEstimatorInterface estimator;

  protected int maxDepth = 2;

  private int nrVariations = 0;
}

